"""
    Copyright (C) 2020 David J. Wooten

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.optimize import curve_fit
from . import musyc_jacobian as musyc_jacobian
from .base import *
from .. import utils
from ..single import hill as hill

class MuSyC(ParameterizedModel):
    """Multidimensional Synergy of Combinations (MuSyC) is a drug synergy framework based on the law of mass action (doi: 10.1016/j.cels.2019.01.003, doi: 10.1101/683433). In MuSyC, synergy is parametrically defined as shifts in potency, efficacy, or cooperativity.

    Synergy Parameters
    ------------------

    alpha21 : Synergistic potency ([0,1) = antagonism, (1,inf) = synergism)
        At large concentrations of drug 2, the "effective dose" of drug 1 = alpha21*d1.
    
    alpha12 : Synergistic potency ([0,1) = antagonism, (1,inf) = synergism)
        At large concentrations of drug 1, the "effective dose" of drug 2 = alpha12*d2.

    beta : Synergistic efficacy ((-inf,0) = antagonism, (0,inf) = synergism)
        At large concentrations of both drugs, the combination achieves an effect beta-% stronger (or weaker) than the stronger single-drug.

    NOTE: The current implementation does not yet fit gamma21 or gamma12, but fixes them at 1.

    gamma21 : Synergistic cooperativity ([0,1) = antagonism, (1,inf) = synergism)
        At large concentrations of drug 2, the Hill slope of drug 1 = gamma21*h1

    gamma12 : Synergistic cooperativity ([0,1) = antagonism, (1,inf) = synergism)
        At large concentrations of drug 1, the Hill slope of drug 2 = gamma12*h2

    """
    def __init__(self, h1_bounds=(0,np.inf), h2_bounds=(0,np.inf),  \
            C1_bounds=(0,np.inf), C2_bounds=(0,np.inf),             \
            E0_bounds=(-np.inf,np.inf), E1_bounds=(-np.inf,np.inf), \
            E2_bounds=(-np.inf,np.inf), E3_bounds=(-np.inf,np.inf), \
            alpha12_bounds=(0,np.inf), alpha21_bounds=(0,np.inf),   \
            gamma12_bounds=(0,np.inf), gamma21_bounds=(0,np.inf),   \
            r1=1., r2=1., E0=None, E1=None, E2=None, E3=None,   \
            h1=None, h2=None, C1=None, C2=None, alpha12=None,       \
            alpha21=None, gamma12=1., gamma21=1.):
        super().__init__()
        self.C1_bounds = C1_bounds
        self.C2_bounds = C2_bounds
        self.h1_bounds = h1_bounds
        self.h2_bounds = h2_bounds
        self.E0_bounds = E0_bounds
        self.E1_bounds = E1_bounds
        self.E2_bounds = E2_bounds
        self.E3_bounds = E3_bounds
        self.alpha12_bounds = alpha12_bounds
        self.alpha21_bounds = alpha21_bounds
        self.gamma12_bounds = gamma12_bounds
        self.gamma21_bounds = gamma21_bounds

        self.r1 = r1
        self.r2 = r2
        self.E0 = E0
        self.E1 = E1
        self.E2 = E2
        self.E3 = E3
        self.h1 = h1
        self.h2 = h2
        self.C1 = C1
        self.C2 = C2
        self.alpha12 = alpha12
        self.alpha21 = alpha21
        self.gamma12 = gamma12
        self.gamma21 = gamma21
        self.beta = None

        self.converged = False

        with np.errstate(divide='ignore'):
            self.logh1_bounds = (np.log(h1_bounds[0]), np.log(h1_bounds[1]))
            self.logC1_bounds = (np.log(C1_bounds[0]), np.log(C1_bounds[1]))
            self.logh2_bounds = (np.log(h2_bounds[0]), np.log(h2_bounds[1]))
            self.logC2_bounds = (np.log(C2_bounds[0]), np.log(C2_bounds[1]))
            
            self.logalpha12_bounds = (np.log(alpha12_bounds[0]), np.log(alpha12_bounds[1]))
            self.logalpha21_bounds = (np.log(alpha21_bounds[0]), np.log(alpha21_bounds[1]))

            self.loggamma12_bounds = (np.log(gamma12_bounds[0]), np.log(gamma12_bounds[1]))
            self.loggamma21_bounds = (np.log(gamma21_bounds[0]), np.log(gamma21_bounds[1]))

    def fit(self, d1, d2, E, drug1_model=None, drug2_model=None, use_jacobian = True, p0=None, **kwargs):
        
        f = lambda d, E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21: self._model(d[0], d[1], E0, E1, E2, E3, np.exp(logh1), np.exp(logh2), np.exp(logC1), np.exp(logC2), self.r1, self.r2, np.exp(logalpha12), np.exp(logalpha21))

        jacobian = lambda d, E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21: musyc_jacobian.jacobian(d[0], d[1], E0, E1, E2, E3, logh1, logh2, logC1, logC2, self.r1, self.r2, logalpha12, logalpha21)

        bounds = tuple(zip(self.E0_bounds, self.E1_bounds, self.E2_bounds, self.E3_bounds, self.logh1_bounds, self.logh2_bounds, self.logC1_bounds, self.logC2_bounds, self.logalpha12_bounds, self.logalpha21_bounds))

        p0 = self._get_intial_guess(d1, d2, E, bounds, drug1_model=drug1_model, drug2_model=drug2_model, p0=p0)

        d1 = utils.remove_zeros(d1)
        d2 = utils.remove_zeros(d2)
        xdata = np.vstack((d1,d2))

        try:
            if (use_jacobian):
                popt1, pcov = curve_fit(f, xdata, E, bounds=bounds, jac=jacobian, p0=p0, **kwargs)

            else:
                popt1, pcov = curve_fit(f, xdata, E, bounds=bounds, p0=p0, **kwargs)

            E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21 = popt1
            self.converged = True
        except RuntimeError:
            E0, E1, E2, E3, logh1, logh2, logC1, logC2, logalpha12, logalpha21 = p0
            self.converged = False

        self.E0 = E0
        self.E1 = E1
        self.E2 = E2
        self.E3 = E3
        self.h1 = np.exp(logh1)
        self.h2 = np.exp(logh2)
        self.C1 = np.exp(logC1)
        self.C2 = np.exp(logC2)
        self.alpha12 = np.exp(logalpha12)
        self.alpha21 = np.exp(logalpha21)
        self.beta = (min(E1,E2)-E3) / (E0 - min(E1,E2))
        self.gamma12 = 1.
        self.gamma21 = 1.

        if (self._is_parameterized()):
            self._score(d1, d2, E)
            
        return self.alpha12, self.alpha21, self.beta

    def _get_intial_guess(self, d1, d2, E, bounds, drug1_model=None, drug2_model=None, p0=None):
        if p0 is not None:
            p0 = list(p0)
            # p0 = (E0, E1, E2, E3, h1, h2, C1, C2, alpha12, alpha21, gamma12, gamma21) - starting from h1 to the end, all guesses must be log-transformed
            for i in range(4,len(p0)):
                p0[i] = np.log(p0[i])
            utils.sanitize_initial_guess(p0, bounds)
            return p0
        else:
            if drug1_model is None:
                mask = np.where(d2==min(d2))
                drug1_model = hill.Hill.create_fit(d1[mask], E[mask], E0_bounds=self.E0_bounds, Emax_bounds=self.E1_bounds, h_bounds=self.h1_bounds, C_bounds=self.C1_bounds)
            if drug2_model is None:
                mask = np.where(d1==min(d1))
                drug2_model = hill.Hill.create_fit(d2[mask], E[mask], E0_bounds=self.E0_bounds, Emax_bounds=self.E2_bounds, h_bounds=self.h2_bounds, C_bounds=self.C2_bounds)
            
            # Get initial guesses of E0, E1, E2, h1, h2, C1, and C2 from single-drug fits
            E0_1, E1, h1, C1 = drug1_model.get_parameters()
            E0_2, E2, h2, C2 = drug2_model.get_parameters()
            
            # Get initial guess of E3 at E(d1_max, d2_max), if that point exists
            E3 = E[(d1==max(d1)) & (d2==max(d2))]
            if len(E3)>0: E3 = np.mean(E3)
            # TODO: Otherwise, guess E3 is the minimum value observed - THIS assumes the orientation of E is decreasing as d increases
            # Could infer orientation from sign of Y=(E0-E1)+(E0-E2). Y>0 means drug causes E to decrease. Y<0 means drug causes E to increase. Or let user set it
            else: E3 = np.min(E)
            
            p0 = [(E0_1+E0_2)/2., E1, E2, E3, np.log(h1), np.log(h2), np.log(C1), np.log(C2), 0, 0] #, 0, 0] # For gamma
            
            utils.sanitize_initial_guess(p0, bounds)

            return p0

    def E(self, d1, d2):
        if not self._is_parameterized():
            return 0
            #raise ModelNotParameterizedError()
        return self._model(d1, d2, self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.r1, self.r2, self.alpha12, self.alpha21)

    def get_parameters(self):
        #return self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.alpha12, self.alpha21, self.gamma12, self.gamma21, self.r1, self.r2
        return self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.alpha12, self.alpha21, self.r1, self.r2


    def _C_to_r1r(self, C, h, r1):
        return r1*C**h

    def _r_to_C(self, h, r1r):
        return (r1r/r1)**(1./h)

    def _model(self, d1, d2, E0, E1, E2, E3, h1, h2, C1, C2, r1, r2, alpha12, alpha21):

        d1h1 = d1**h1
        d2h2 = d2**h2
        alpha21d1h1 = (alpha21*d1)**h1
        alpha12d2h2 = (alpha12*d2)**h2
        r1r = r1*C1**h1
        r2r = r2*C2**h2

        U = r1r*r2r*(r1*(alpha21d1h1) + r1r + r2*(alpha12d2h2) + r2r)/(d1h1*r1**2*r2*(alpha12d2h2)*(alpha21d1h1) + d1h1*r1**2*r2r*(alpha21d1h1) + d1h1*r1*r1r*r2*(alpha12d2h2) + d1h1*r1*r1r*r2r + d1h1*r1*r2*r2r*(alpha12d2h2) + d1h1*r1*r2r**2 + d2h2*r1*r1r*r2*(alpha21d1h1) + d2h2*r1*r2**2*(alpha12d2h2)*(alpha21d1h1) + d2h2*r1*r2*r2r*(alpha21d1h1) + d2h2*r1r**2*r2 + d2h2*r1r*r2**2*(alpha12d2h2) + d2h2*r1r*r2*r2r + r1*r1r*r2r*(alpha21d1h1) + r1r**2*r2r + r1r*r2*r2r*(alpha12d2h2) + r1r*r2r**2)

        A1 = r1*r2r*(d1h1*r1*(alpha21d1h1) + d1h1*r1r + d1h1*r2r + d2h2*r2*(alpha21d1h1))/(d1h1*r1**2*r2*(alpha12d2h2)*(alpha21d1h1) + d1h1*r1**2*r2r*(alpha21d1h1) + d1h1*r1*r1r*r2*(alpha12d2h2) + d1h1*r1*r1r*r2r + d1h1*r1*r2*r2r*(alpha12d2h2) + d1h1*r1*r2r**2 + d2h2*r1*r1r*r2*(alpha21d1h1) + d2h2*r1*r2**2*(alpha12d2h2)*(alpha21d1h1) + d2h2*r1*r2*r2r*(alpha21d1h1) + d2h2*r1r**2*r2 + d2h2*r1r*r2**2*(alpha12d2h2) + d2h2*r1r*r2*r2r + r1*r1r*r2r*(alpha21d1h1) + r1r**2*r2r + r1r*r2*r2r*(alpha12d2h2) + r1r*r2r**2)

        A2 = r1r*r2*(d1h1*r1*(alpha12d2h2) + d2h2*r1r + d2h2*r2*(alpha12d2h2) + d2h2*r2r)/(d1h1*r1**2*r2*(alpha12d2h2)*(alpha21d1h1) + d1h1*r1**2*r2r*(alpha21d1h1) + d1h1*r1*r1r*r2*(alpha12d2h2) + d1h1*r1*r1r*r2r + d1h1*r1*r2*r2r*(alpha12d2h2) + d1h1*r1*r2r**2 + d2h2*r1*r1r*r2*(alpha21d1h1) + d2h2*r1*r2**2*(alpha12d2h2)*(alpha21d1h1) + d2h2*r1*r2*r2r*(alpha21d1h1) + d2h2*r1r**2*r2 + d2h2*r1r*r2**2*(alpha12d2h2) + d2h2*r1r*r2*r2r + r1*r1r*r2r*(alpha21d1h1) + r1r**2*r2r + r1r*r2*r2r*(alpha12d2h2) + r1r*r2r**2)

        #A12 = r1*r2*(d1h1*r1*(alpha12d2h2)*(alpha21d1h1) + d1h1*r2r*(alpha12d2h2) + d2h2*r1r*(alpha21d1h1) + d2h2*r2*(alpha12d2h2)*(alpha21d1h1))/(d1h1*r1**2*r2*(alpha12d2h2)*(alpha21d1h1) + d1h1*r1**2*r2r*(alpha21d1h1) + d1h1*r1*r1r*r2*(alpha12d2h2) + d1h1*r1*r1r*r2r + d1h1*r1*r2*r2r*(alpha12d2h2) + d1h1*r1*r2r**2 + d2h2*r1*r1r*r2*(alpha21d1h1) + d2h2*r1*r2**2*(alpha12d2h2)*(alpha21d1h1) + d2h2*r1*r2*r2r*(alpha21d1h1) + d2h2*r1r**2*r2 + d2h2*r1r*r2**2*(alpha12d2h2) + d2h2*r1r*r2*r2r + r1*r1r*r2r*(alpha21d1h1) + r1r**2*r2r + r1r*r2*r2r*(alpha12d2h2) + r1r*r2r**2)

        #A12 = 1-(U+A1+A2)
        
        return U*E0 + A1*E1 + A2*E2 + (1-(U+A1+A2))*E3

    def create_fit(d1, d2, E, h1_bounds=(0,np.inf), h2_bounds=(0,np.inf),  \
            C1_bounds=(0,np.inf), C2_bounds=(0,np.inf),             \
            E0_bounds=(-np.inf,np.inf), E1_bounds=(-np.inf,np.inf), \
            E2_bounds=(-np.inf,np.inf), E3_bounds=(-np.inf,np.inf), \
            alpha12_bounds=(0,np.inf), alpha21_bounds=(0,np.inf),   \
            gamma12_bounds=(0,np.inf), gamma21_bounds=(0,np.inf),   \
            r1=1., r2=1., **kwargs):

        model = MuSyC(E0_bounds=E0_bounds, E1_bounds=E1_bounds, E2_bounds=E2_bounds, E3_bounds=E3_bounds, h1_bounds=h1_bounds, h2_bounds=h2_bounds, C1_bounds=C1_bounds, C2_bounds=C2_bounds, alpha12_bounds=alpha12_bounds, alpha21_bounds=alpha21_bounds, gamma12_bounds=gamma12_bounds, gamma21_bounds=gamma21_bounds, r1=r1, r2=r2)
        model.fit(d, E, **kwargs)
        return model

    def __repr__(self):
        if not self._is_parameterized(): return "MuSyC()"
        return "MuSyC(E0=%0.2f, E1=%0.2f, E2=%0.2f, E3=%0.2f, h1=%0.2f, h2=%0.2f, C1=%0.2e, C2=%0.2e, r1=%0.2f, r2=%0.2f, alpha12=%0.2f, alpha21=%0.2f, beta=%0.2f, converged=%r)"%(self.E0, self.E1, self.E2, self.E3, self.h1, self.h2, self.C1, self.C2, self.r1, self.r2, self.alpha12, self.alpha21, self.beta, self.converged)


"""
    Copyright (C) 2020 David J. Wooten

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
try:
    import pandas as pd
except ModuleNotFoundError:
    pass

def to_synergyfinder(d1, d2, E, d1_name="drug1", d2_name="drug2", d1_unit="uM", d2_unit="uM"):
    return pd.DataFrame(dict(block_id=1, drug_col=d1_name, drug_row=d2_name, conc_c=d1, conc_r=d2, response=E, conc_c_unit=d1_unit, conc_r_unit=d2_unit))




"""
    Copyright (C) 2020 David J. Wooten

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import warnings
from scipy.optimize import curve_fit
from .. import utils
from ..single import hill as hill
from .base import *

class Zimmer(ParameterizedModel):
    """The Effective Dose Model from Zimmer et al (doi: 10.1073/pnas.1606301113). This model uses the multiplicative survival principle (i.e., Bliss), but adds a parameter for each drug describing how it affects the potency of the other. Specifically, given doses d1 and d2, this model translates them to "effective" doses using the following system of equations

                            d1
    d1_eff =  --------------------------------
              1 + a12*(1/(1+(d2_eff/C2)^(-1)))

                            d2
    d2_eff =  --------------------------------
              1 + a21*(1/(1+(d1_eff/C1)^(-1)))

    Synergy Parameters
    ------------------

    a12 : (-(1+(d2_eff/C2))/(d2_eff/C2),0)=synergism, (0,inf)=antagonism
        Describes how drug 2 affects the effective dose of drug 1.

    a21 : (-(1+(d1_eff/C1))/(d1_eff/C1),0)=synergism, (0,inf)=antagonism
        Describes how drug 1 affects the effective dose of drug 2.
    """
    def __init__(self, h1_bounds=(0,np.inf), h2_bounds=(0,np.inf),  \
            C1_bounds=(0,np.inf), C2_bounds=(0,np.inf), a12_bounds=(-np.inf, np.inf), a21_bounds=(-np.inf, np.inf), h1=None, h2=None, C1=None, C2=None, a12=None, a21=None):

        super().__init__()
        self.C1_bounds = C1_bounds
        self.C2_bounds = C2_bounds
        self.h1_bounds = h1_bounds
        self.h2_bounds = h2_bounds
        self.a12_bounds = a12_bounds
        self.a21_bounds = a21_bounds

        with np.errstate(divide='ignore'):
            self.logh1_bounds = (np.log(h1_bounds[0]), np.log(h1_bounds[1]))
            self.logC1_bounds = (np.log(C1_bounds[0]), np.log(C1_bounds[1]))
            self.logh2_bounds = (np.log(h2_bounds[0]), np.log(h2_bounds[1]))
            self.logC2_bounds = (np.log(C2_bounds[0]), np.log(C2_bounds[1]))

        self.h1 = h1
        self.h2 = h2
        self.C1 = C1
        self.C2 = C2
        self.a12 = a12
        self.a21 = a21

    def fit(self, d1, d2, E, drug1_model=None, drug2_model=None, use_jacobian=True, **kwargs):
        if (use_jacobian):
            warnings.warn(FeatureNotImplemented("Jacobian has not been implemented for Zimmer synergy model, but will still be used for single-drug fits"))
        bounds = tuple(zip(self.logh1_bounds, self.logh2_bounds, self.logC1_bounds, self.logC2_bounds, self.a12_bounds, self.a21_bounds))

        if 'p0' in kwargs:
            p0 = list(kwargs.get('p0'))
            for i in range(4):
                p0[i] = np.log(p0[i])
            utils.sanitize_initial_guess(p0, bounds)
            kwargs['p0'] = p0
        else:
            if drug1_model is None:
                mask = np.where(d2==min(d2))
                drug1_model = hill.Hill_2P.create_fit(d1[mask], E[mask], h_bounds=self.h1_bounds, C_bounds=self.C1_bounds, use_jacobian=use_jacobian)
                
            if drug2_model is None:
                mask = np.where(d1==min(d1))
                drug2_model = hill.Hill_2P.create_fit(d2[mask], E[mask], h_bounds=self.h2_bounds, C_bounds=self.C2_bounds, use_jacobian=use_jacobian)
            
            E0_1, E1, h1, C1 = drug1_model.get_parameters()
            E0_2, E2, h2, C2 = drug2_model.get_parameters()
            p0 = [h1, h2, C1, C2, 0, 0]
            for i in range(4):
                p0[i] = np.log(p0[i])
            utils.sanitize_initial_guess(p0, bounds)
            kwargs['p0']  = p0


        xdata = np.vstack((d1,d2))
        
        f = lambda d, logh1, logh2, logC1, logC2, a12, a21: self._model(d[0], d[1], np.exp(logh1), np.exp(logh2), np.exp(logC1), np.exp(logC2), a12, a21)
        
        with np.errstate(divide='ignore', invalid='ignore'):
            popt1, pcov = curve_fit(f, xdata, E, bounds=bounds, **kwargs)

        logh1, logh2, logC1, logC2, a12, a21 = popt1
        self.h1 = np.exp(logh1)
        self.h2 = np.exp(logh2)
        self.C1 = np.exp(logC1)
        self.C2 = np.exp(logC2)
        self.a12 = a12
        self.a21 = a21

        self._score(d1, d2, E)

        return a12, a21

    def E(self, d1, d2):
        if not self._is_parameterized():
            # ERROR
            return 0
        return self._model(d1, d2, self.h1, self.h2, self.C1, self.C2, self.a12, self.a21)

    def _model(self, d1, d2, h1, h2, C1, C2, a12, a21):
        A = d2 + C2*(a21+1) + d2*a12
        B = d2*C1 + C1*C2 + a12*d2*C1 - d1*(d2+C2*(a21+1))
        C = -d1*(d2*C1 + C1*C2)

        d1p = (-B + np.sqrt(np.power(B,2.) - 4*A*C)) / (2.*A)
        d2p = d2 / (1. + a21 / (1. + C1/d1p))
        
        return (1 - d1p**h1/(C1**h1+d1p**h1)) * (1 - d2p**h2/(C2**h2+d2p**h2))

    def get_parameters(self):
        return self.h1, self.h2, self.C1, self.C2, self.a12, self.a21

    def create_fit(d1, d2, E, h1_bounds=(0,np.inf), h2_bounds=(0,np.inf), C1_bounds=(0,np.inf), C2_bounds=(0,np.inf), a12_bounds=(-np.inf,np.inf), a21_bounds=(-np.inf,np.inf), **kwargs):
        model = Zimmer(h1_bounds=h1_bounds, h2_bounds=h2_bounds, C1_bounds=C1_bounds, C2_bounds=C2_bounds, a12_bounds=a12_bounds, a21_bounds=a21_bounds)
        
        model.fit(d, E, **kwargs)
        return model

    def __repr__(self):
        if not self._is_parameterized(): return "Zimmer()"
        return "Zimmer(h1=%0.2f, h2=%0.2f, C1=%0.2e, C2=%0.2e, a12=%0.2f, a21=%0.2f)"%(self.h1, self.h2, self.C1, self.C2, self.a12, self.a21)


"""
    Copyright (C) 2020 David J. Wooten

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from .. import utils
from ..utils import plots


class ParameterizedModel:
    """
    This is the base class for paramterized synergy models, including MuSyC, Zimmer, GPDI, and BRAID.
    """
    def __init__(self):
        """Bounds for drug response parameters (for instance, given percent viability data, one might expect E to be bounded within (0,1)) can be set, or parameters can be explicitly set.
        """

        self.sum_of_squares_residuals = None
        self.r_squared = None
        self.aic = None
        self.bic = None
#        self.converged = False

    def _score(self, d1, d2, E):
        """Calculate goodness of fit and model quality scores, including sum-of-squares residuals, R^2, Akaike Information Criterion (AIC), and Bayesian Information Criterion (BIC).

        If model is not yet paramterized, does nothing

        Called automatically during model.fit(d1, d2, E)

        Parameters
        ----------
        d1 : array-like
            Doses of drug 1
        
        d2 : array-like
            Doses of drug 2
        
        E : array-like
            Dose-response at doses d1 and d2
        """
        if (self._is_parameterized()):

            n_parameters = len(self.get_parameters())

            self.sum_of_squares_residuals = utils.residual_ss(d1, d2, E, self.E)
            self.r_squared = utils.r_squared(E, self.sum_of_squares_residuals)
            self.aic = utils.AIC(self.sum_of_squares_residuals, n_parameters, len(E))
            self.bic = utils.BIC(self.sum_of_squares_residuals, n_parameters, len(E))

    def get_parameters():
        """
        Returns
        ----------
        parameters : list or tuple
            Model's parameters
        """
        return []

    def fit(self, d1, d2, E, drug1_model=None, drug2_model=None, use_jacobian = True, p0=None, **kwargs):
        """Fit the synergy model to data using scipy.optimize.curve_fit().

        Parameters
        ----------
        d1 : array-like
            Doses of drug 1
        
        d2 : array-like
            Doses of drug 2

        E : array-like
            Dose-response at doses d1 and d2

        drug1_model : single-drug-model, default=None
            Only used when p0 is None. Pre-defined, or fit, model (e.g., Hill()) of drug 1 alone. Parameters from this model are used to provide an initial guess of E0, E1, h1, and C1 for the 2D-model fit. If None (and p0 is None), then d1 and E will be masked where d2==min(d2), and used to fit a model for drug 1.

        drug2_model : single-drug-model, default=None
            Same as drug1_model, for drug 2.
        
        use_jacobian : bool, default=True
            If True, will use the Jacobian to help guide fit (ONLY MuSyC, Hill, and Hill_2P have Jacobian implemented yet). When the number
            of data points is less than a few hundred, this makes the fitting
            slower. However, it also improves the reliability with which a fit
            can be found. If drug1_model or drug2_model are None, use_jacobian will also be applied for their fits.

        p0 : tuple, default=None
            Initial guess for the parameters. If p0 is None (default), drug1_model and drug2_model will be used to obtain an initial guess. If they are also None, they will be fit to the data. If they fail to fit, the initial guess will be E0=max(E), Emax=min(E), h=1, C=median(d), and all synergy parameters are additive (i.e., at the boundary between antagonistic and synergistic)
        
        kwargs
            kwargs to pass to scipy.optimize.curve_fit()

        Returns
        ----------
        synergy_parameters : array-like
            The fit parameters describing the synergy in the data
        """

    def E(self, d1, d2):
        """
        Parameters
        ----------
        d1 : array-like
            Doses of drug 1
        
        d2 : array-like
            Doses of drug 2
        
        Returns
        ----------
        effect : array-like
            Evaluate's the model at doses d1 and d2
        """
        return 0

    def _is_parameterized(self):
        """
        Returns
        ----------
        is_parameterzed : bool
            True if all of the parameters are set. False if any are None or nan.
        """
        return not (None in self.get_parameters() or True in np.isnan(np.asarray(self.get_parameters())))

    def plot_colormap(self, d1, d2, cmap="viridis", **kwargs):
        """Plots the model's effect, E(d1, d2) as a heatmap

        Parameters
        ----------
        d1 : array-like
            Doses of drug 1
        
        d2 : array-like
            Doses of drug 2
        
        kwargs
            kwargs passed to synergy.utils.plots.plot_colormap()
        """
        if not self._is_parameterized():
            #raise ModelNotParameterizedError()
            return
        
        E = self.E(d1, d2)
        plots.plot_colormap(d1, d2, E, cmap=cmap, **kwargs)

    def plot_surface_plotly(self, d1, d2, cmap="RdBu", **kwargs):
        """Plots the model's effect, E(d1, d2) as a surface using synergy.utils.plots.plot_surface_plotly()

        Parameters
        ----------
        d1 : array-like
            Doses of drug 1
        
        d2 : array-like
            Doses of drug 2
        
        kwargs
            kwargs passed to synergy.utils.plots.plot_colormap()
        """
        if not self._is_parameterized():
            #raise ModelNotParameterizedError()
            return
        
        E = self.E(d1, d2)
        plots.plot_surface_plotly(d1, d2, E, cmap=cmap, **kwargs)

    def create_fit(d1, d2, E, **kwargs):
        """Courtesy one-liner to create a model and fit it to data. Appropriate (see model __init__() for details) bounds may be set for curve_fit.

        Parameters
        ----------
        d1 : array-like
            Doses of drug 1
        
        d2 : array-like
            Doses of drug 2
        
        E : array-like
            Dose-response at doses d1 and d2

        X_bounds : tuple, 
            Bounds for each parameter of the model to be fit. See model.__init__() for specific details.
        
        kwargs
            kwargs to pass sto model.fit()

        Returns
        ----------
        model : ParametricModel
            Synergy model fit to the given data
        """
        return


class DoseDependentModel:
    def __init__(self, h1_bounds=(0,np.inf), h2_bounds=(0,np.inf),  \
            C1_bounds=(0,np.inf), C2_bounds=(0,np.inf),             \
            E0_bounds=(-np.inf,np.inf), E1_bounds=(-np.inf,np.inf), \
            E2_bounds=(-np.inf,np.inf)):
        """
        Parameters
        ----------
        E0_bounds: tuple, default=(-np.inf, np.inf)
            Bounds to use for E0 to fit drug1_model and drug2_model if they are not supplied directly in .fit()
        
        E{X}_bounds: tuple, default=(-np.inf, np.inf)
            Bounds to use for Emax to fit drug{X}_model if it is not supplied directly in .fit() (e.g., E1_bounds will constrain drug 1's Emax)
        
        h{X}_bounds: tuple, default=(0, np.inf)
            Bounds to use for hill slope to fit drug{X}_model if it is not supplied directly in .fit()
        
        C{X}_bounds: tuple, default=(0, np.inf)
            Bounds to use for EC50 to fit drug{X}_model if it is not supplied directly in .fit()
        """
        self.C1_bounds = C1_bounds
        self.C2_bounds = C2_bounds
        self.h1_bounds = h1_bounds
        self.h2_bounds = h2_bounds
        self.E0_bounds = E0_bounds
        self.E1_bounds = E1_bounds
        self.E2_bounds = E2_bounds

        self.synergy = None
        self.d1 = None
        self.d2 = None

        self.drug1_model = None
        self.drug2_model = None

    def fit(self, d1, d2, E, drug1_model=None, drug2_model=None, **kwargs):
        """Calculates dose-dependent synergy at doses d1, d2.

        Parameters
        ----------
        d1 : array-like
            Doses of drug 1
        
        d2 : array-like
            Doses of drug 2

        E : array-like
            Dose-response at doses d1 and d2

        drug1_model : single-drug-model, default=None
            Pre-defined, or fit, model (e.g., Hill()) of drug 1 alone. If None (default), then d1 and E will be masked where d2==min(d2), and used to fit a model (Hill, Hill_2P, or Hill_CI, depending on the synergy model) for drug 1.

        drug2_model : single-drug-model, default=None
            Same as drug1_model, for drug 2.
        
        kwargs
            kwargs to pass to Hill.fit() (or whichever single-drug model is used)

        Returns
        ----------
        synergy : array-like
            The synergy calculated at all doses d1, d2
        """
        self.d1 = d1
        self.d2 = d2
        self.synergy = 0*d1
        self.synergy[:] = np.nan
        return self.synergy

    def plot_colormap(self, cmap="PRGn", neglog=False, **kwargs):
        """Plots the synergy as a heatmap

        Parameters
        ----------
        cmap : string, default="PRGn"
            Colorscale for the plot

        neglog : bool, default=False
            If True, will transform the synergy values by -log(synergy). Loewe and CI are synergistic between [0,1) and antagonistic between (1,inf). Thus, -log(synergy) becomes synergistic for positive values, and antagonistic for negative values. This behavior matches other synergy frameworks, and helps better visualize results. But it is never set by default.

        kwargs
            kwargs passed to synergy.utils.plots.plot_colormap()
        """
        if neglog:
            with np.errstate(invalid="ignore"):
                plots.plot_colormap(self.d1, self.d2, -np.log(self.synergy), cmap=cmap, **kwargs)
        else:
            plots.plot_colormap(self.d1, self.d2, self.synergy, cmap=cmap, **kwargs)

    def plot_surface_plotly(self, cmap="PRGn", **kwargs):
        """Plots the synergy as a 3D surface using synergy.utils.plots.plot_surface_plotly()

        Parameters
        ----------
        kwargs
            kwargs passed to synergy.utils.plots.plot_colormap()
        """
        plots.plot_surface_plotly(self.d1, self.d2, self.synergy, cmap=cmap, **kwargs)

class ModelNotParameterizedError(Exception):
    """
    The model must be parameterized prior to use. This can be done by calling
    fit(), or setParameters().
    """
    def __init__(self, msg='The model must be parameterized prior to use. This can be done by calling fit(), or setParameters().', *args, **kwargs):
        super().__init__(msg, *args, **kwargs)

class FeatureNotImplemented(Warning):
    """
    """
    def __init__(self, msg="This feature is not yet implemented", *args, **kwargs):
        super().__init__(msg, *args, **kwargs)



"""
    Copyright (C) 2020 David J. Wooten

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from scipy.optimize import curve_fit
from scipy.stats import linregress
import numpy as np
import synergy.utils.utils as utils

class Hill:
    """The four-parameter Hill equation

                            d^h
    E = E0 + (Emax-E0) * ---------
                         C^h + d^h

    The Hill equation is a standard model for single-drug dose-response curves.
    This is the base model for Hill_2P and Hill_CI.
    """
    def __init__(self, E0=None, Emax=None, h=None, C=None, E0_bounds=(-np.inf, np.inf), Emax_bounds=(-np.inf, np.inf), h_bounds=(0,np.inf), C_bounds=(0,np.inf)):
        """
        Parameters
        ----------
        E0 : float, optional
            Effect at 0 dose. Set this if you are creating a synthetic Hill
            model, rather than fitting from data
        
        Emax : float, optional
            Effect at 0 dose. Set this if you are creating a synthetic Hill
            model, rather than fitting from data
        
        h : float, optional
            The Hill-slope. Set this if you are creating a synthetic Hill
            model, rather than fitting from data
        
        C : float, optional
            EC50, the dose for which E = (E0+Emax)/2. Set this if you are
            creating a synthetic Hill model, rather than fitting from data
        
        E0_bounds: tuple, default=(-np.inf, np.inf)
            Bounds to use for E0 when fitting this model to data
        
        Emax_bounds: tuple, default=(-np.inf, np.inf)
            Bounds to use for Emax when fitting this model to data
        
        h_bounds: tuple, default=(0, np.inf)
            Bounds to use for h when fitting this model to data
        
        C_bounds: tuple, default=(0, np.inf)
            Bounds to use for C when fitting this model to data
        """
        self.E0 = E0
        self.Emax = Emax
        self.h = h
        self.C = C
        
        self.E0_bounds=E0_bounds
        self.Emax_bounds=Emax_bounds
        self.h_bounds=h_bounds
        self.C_bounds=C_bounds
        with np.errstate(divide='ignore'):
            self.logh_bounds = (np.log(h_bounds[0]), np.log(h_bounds[1]))
            self.logC_bounds = (np.log(C_bounds[0]), np.log(C_bounds[1]))

        self.converged = False
    
    def fit(self, d, E, use_jacobian=True, **kwargs):
        """Fit the Hill equation to data. Fitting algorithm searches for h and C in a log-scale, but all bounds and guesses should be provided in a linear scale.

        Parameters
        ----------
        d : array-like
            Array of doses measured
        
        E : array-like
            Array of effects measured at doses d
        
        use_jacobian : bool, default=True
            If True, will use the Jacobian to help guide fit. When the number
            of data points is less than a few hundred, this makes the fitting
            slower. However, it also improves the reliability with which a fit
            can be found.
        
        kwargs
            kwargs to pass to scipy.optimize.curve_fit()
        """
        f = lambda d, E0, E1, logh, logC: self._model(d, E0, E1, np.exp(logh), np.exp(logC))

        bounds = tuple(zip(self.E0_bounds, self.Emax_bounds, self.logh_bounds, self.logC_bounds))

        d = utils.remove_zeros(d)
        

        if 'p0' in kwargs:
            p0 = list(kwargs.get('p0'))
            p0[2] = np.log(p0[2])
            p0[3] = np.log(p0[3])
            utils.sanitize_initial_guess(p0, bounds)
            kwargs['p0'] = p0
        else:
            p0 = [max(E), min(E), 0, np.log(np.median(d))]
            utils.sanitize_initial_guess(p0, bounds)
            kwargs['p0'] = p0

        with np.errstate(divide='ignore', invalid='ignore'):
            try:
                if use_jacobian:
                    popt1, pcov = curve_fit(f, d, E, bounds=bounds, jac=self._model_jacobian, **kwargs)
                else: 
                    popt1, pcov = curve_fit(f, d, E, bounds=bounds, **kwargs)
                E0, E1, logh, logC = popt1
                self.converged = True
            except RuntimeError:
                #print("\n\n*********\nFailed to fit single drug\n*********\n\n")
                E0 = np.max(E)
                E1 = np.min(E)
                logh = 0
                logC = np.log(np.median(d))
                self.converged = False

        self.E0 = E0
        self.Emax = E1
        self.h = np.exp(logh)
        self.C = np.exp(logC)

    def E(self, d):
        """Evaluate this model at dose d. If the model is not parameterized, returns 0.

        Parameters
        ----------
        d : array-like
            Doses to calculate effect at
        
        Returns
        ----------
        effect : array-like
            Evaluate's the model at dose in d
        """
        if not self._is_parameterized():
            return 0
        return self._model(d, self.E0, self.Emax, self.h, self.C)

    def E_inv(self, E):
        """Inverse of the Hill equation

        Parameters
        ----------
        E : array-like
            Effects to get the doses for
        
        Returns
        ----------
        doses : array-like
            Doses which achieve effects E using this model. Effects that are
            outside the range [E0, Emax] will return np.nan for the dose
        """
        if not self._is_parameterized():
            return 0
        return self._model_inv(E, self.E0, self.Emax, self.h, self.C)

    def get_parameters(self):
        """
        Returns
        ----------
        parameters : tuple
            (E0, Emax, h, C)
        """
        return (self.E0, self.Emax, self.h, self.C)
        
    def _model(self, d, E0, Emax, h, C):
        dh = d**h
        return E0 + (Emax-E0)*dh/(C**h+dh)

    def _model_inv(self, E, E0, Emax, h, C):
        d = np.float_power((E0-Emax)/(E-Emax)-1.,1./h)*C
        return d

    def _model_jacobian(self, d, E0, Emax, logh, logC):
        """
        Returns
        ----------
        jacobian : array-like
            Derivatives of the Hill equation with respect to E0, Emax, logh,
            and logC
        """
        
        dh = d**(np.exp(logh))
        Ch = (np.exp(logC))**(np.exp(logh))
        logd = np.log(d)

        jE0 = 1 - dh/(Ch+dh)
        jEmax = 1-jE0

        jC = (E0-Emax)*dh*np.exp(logh+logC)*(np.exp(logC))**(np.exp(logh)-1) / ((Ch+dh)*(Ch+dh))

        jh = (Emax-E0)*dh*np.exp(logh) * ((Ch+dh)*logd - (logC*Ch + logd*dh)) / ((Ch+dh)*(Ch+dh))
        
        return np.hstack((jE0.reshape(-1,1), jEmax.reshape(-1,1), jh.reshape(-1,1), jC.reshape(-1,1)))

    def _is_parameterized(self):
        return None not in (self.E0, self.Emax, self.h, self.C)

    def create_fit(d, E, E0_bounds=(-np.inf, np.inf), Emax_bounds=(-np.inf, np.inf), h_bounds=(0,np.inf), C_bounds=(0,np.inf), **kwargs):
        """Courtesy function to build a Hill model directly from data.
        Initializes a model using the provided bounds, then fits.
        """
        drug = Hill(E0_bounds=E0_bounds, Emax_bounds=Emax_bounds, h_bounds=h_bounds, C_bounds=C_bounds)
        drug.fit(d, E, **kwargs)
        return drug

    def __repr__(self):
        if not self._is_parameterized(): return "Hill()"
        
        return "Hill(E0=%0.2f, Emax=%0.2f, h=%0.2f, C=%0.2e)"%(self.E0, self.Emax, self.h, self.C)

class Hill_2P(Hill):
    """The two-parameter Hill equation

                            d^h
    E = E0 + (Emax-E0) * ---------
                         C^h + d^h

    Mathematically equivalent to the four-parameter Hill equation, but E0 and Emax are held constant (not fit to data).
    """
    def __init__(self, h=None, C=None, h_bounds=(0,np.inf), C_bounds=(0,np.inf), E0=1, Emax=0):
        super().__init__(h=h, C=C, E0=E0, Emax=Emax, h_bounds=h_bounds, C_bounds=C_bounds)

    def fit(self, d, E, use_jacobian=True, **kwargs):
        if self.E0 is None: self.E0 = 1.
        if self.Emax is None: self.Emax = 0.

        f = lambda d, logh, logC: self._model(d, self.E0, self.Emax, np.exp(logh), np.exp(logC))

        d = utils.remove_zeros(d)

        bounds = tuple(zip(self.logh_bounds, self.logC_bounds))

        if 'p0' in kwargs:
            p0 = list(kwargs.get('p0'))
            p0[0] = np.log(p0[0])
            p0[1] = np.log(p0[1])
            kwargs['p0'] = p0
            utils.sanitize_initial_guess(p0, bounds)
        else:
            p0 = [0, np.log(np.median(d))]
            utils.sanitize_initial_guess(p0, bounds)
            kwargs['p0'] = p0

        with np.errstate(divide='ignore', invalid='ignore'):
            try:
                if use_jacobian:
                    popt1, pcov = curve_fit(f, d, E, bounds=bounds, jac=self._model_jacobian, **kwargs)
                else: 
                    popt1, pcov = curve_fit(f, d, E, bounds=bounds, **kwargs)
                logh, logC = popt1
                self.converged = True
            except RuntimeError:
                #print("\n\n*********\nFailed to fit single drug\n*********\n\n")
                logh = 0
                logC = np.log(np.median(d))
                self.converged = False

        self.h = np.exp(logh)
        self.C = np.exp(logC)

    def _model_jacobian(self, d, logh, logC):
        dh = d**(np.exp(logh))
        Ch = (np.exp(logC))**(np.exp(logh))
        logd = np.log(d)
        E0 = self.E0
        Emax = self.Emax

        jC = (E0-Emax)*dh*np.exp(logh+logC)*(np.exp(logC))**(np.exp(logh)-1) / ((Ch+dh)*(Ch+dh))

        jh = (Emax-E0)*dh*np.exp(logh) * ((Ch+dh)*logd - (logC*Ch + logd*dh)) / ((Ch+dh)*(Ch+dh))
        
        return np.hstack((jh.reshape(-1,1), jC.reshape(-1,1)))

    def create_fit(d, E, E0=1, Emax=0, h_bounds=(0,np.inf), C_bounds=(0,np.inf), **kwargs):
        drug = Hill_2P(E0=E0, Emax=Emax, h_bounds=h_bounds, C_bounds=C_bounds)
        drug.fit(d, E, **kwargs)
        return drug

    def __repr__(self):
        if not self._is_parameterized(): return "Hill()"
        
        return "Hill2P(E0=%0.2f, Emax=%0.2f, h=%0.2f, C=%0.2e, converged=%r)"%(self.E0, self.Emax, self.h, self.C)
    
    
class Hill_CI(Hill_2P):
    """Mathematically equivalent two-parameter Hill equation with E0=1 and Emax=0. However, Hill_CI.fit() uses the log-linearization approach to dose-response fitting used by the Combination Index.
    """
    def __init__(self, h=None, C=None):
        super().__init__(h=h, C=C, E0=1., Emax=0.)

    def fit(self, d, E):
        mask = np.where((E < 1) & (E > 0) & (d > 0))
        E = E[mask]
        d = d[mask]
        fU = E
        fA = 1-E

        median_effect_line = linregress(np.log(d),np.log(fA/fU))
        self.h = median_effect_line.slope
        self.C = np.exp(-median_effect_line.intercept / self.h)

    def create_fit(d, E):
        drug = Hill_CI()
        drug.fit(d, E)
        return drug

    def __repr__(self):
        if not self._is_parameterized(): return "Hill_CI()"
        
        return "Hill_CI(h=%0.2f, C=%0.2e)"%(self.h, self.C)
    
    